local HeadCount = LibStub("AceAddon-3.0"):GetAddon("HeadCount2")
local Tracker = HeadCount:NewModule("RaidTracker","AceEvent-3.0","AceTimer-3.0")
if not HeadCount then return end

local L = LibStub("AceLocale-3.0"):GetLocale("HeadCount2")
local BD = HeadCount:GetModule("BossDetection")

function Tracker:OnInitialize()
	-- Setup Event Handlers
	self.isEnabled = true
	self.isWaitlistAcceptanceEnabled = false
	self.isCurrentRaidActive = false
	self.currentBossEvent = HeadCount:NewInstance("BossEvent")
	self.raidListWrapper = HeadCount:NewInstance("RaidListWrapper")
	self.messageHandler = HeadCount:GetModule("MessageHandler")
	BD.RegisterCallback(self,"BossKilled")
	BD.RegisterCallback(self,"BossWipe")
	BD.RegisterCallback(self,"BossEngaged")
	--self:RegisterEvent("RAID_ROSTER_UPDATE")
end


function Tracker:GetInstaceBossList()
	local lockTimeleft, isPreviousInstance, encountersTotal, encountersComplete = GetInstanceLockTimeRemaining();
	for i=1,encountersTotal do
		local bossName, texture, isKilled = GetInstanceLockTimeRemainingEncounter(i);
		InstanceBossList[bossName] = i
	end
end

function Tracker:ResetWrapper()
	self.raidListWrapper = HeadCount:NewInstance("RaidListWrapper")
	self:saveRaidListWrapper()
end

function Tracker:SetRaidListWrapper(wrapper)
	if wrapper then
		self.raidListWrapper = wrapper
		--self:saveRaidListWrapper()
	end
end

function Tracker:saveRaidListWrapper()
	--if self.raidListWrapper then
	--	HeadCount.db.profile.raidListWrapper = self.raidListWrapper
	--end
end

function Tracker:recover()
	if self.isCurrentRaidActive then return end
	local raid = self.raidListWrapper:retrieveMostRecentRaid()
	if raid then
		local ts = HeadCount:getUTCDateTimeInSeconds()
		local activityTime = HeadCount:NewInstance("Time",ts)
		-- better raid check test
		local rc = GetNumRaidMembers()
		if UnitInRaid("player")	or rc > 0 then
			-- owner is present in raid
			if raid.isFinalized  then
				-- most recent raid is finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.raid.final"], HeadCount.TITLE, HeadCount.VERSION))
				self.isCurrentRaidActive = false
			else
				-- most recent raid is NOT finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.raid.nofinal"], HeadCount.TITLE, HeadCount.VERSION))
				raid:recover(activityTime)
				-- Set proper events and flags for recovered raid
				self:beginRaid()
				-- Reload Raid State
				self.currentBossEvent.lastBoss = raid.lastBoss
				self.isCurrentRaidActive = true
			end
		else
			-- owner is NOT present in raid
			if raid.isFinalized  then
				-- most recent raid is finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.noraid.final"], HeadCount.TITLE, HeadCount.VERSION))
				self.isCurrentRaidActive = false
			else
				-- most recent raid is NOT finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.noraid.nofinal"], HeadCount.TITLE, HeadCount.VERSION))
				self:endRaid()
				raid:finalize(activityTime)
				self.isCurrentRaidActive = false
			end
		end
	else
		HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.norecover"],HeadCount.TITLE, HeadCount.VERSION))
	end
end

function Tracker:beginRaid()
	self.isCurrentRaidActive = true
	local waitlistDuration = HeadCount.db.profile.raid.waitlist.duration
	if waitlistDuration > 0 then
		self.isWaitlistAcceptanceEnabled = false
	else
		self.isWaitlistAcceptanceEnabled = true
	end
	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", Tracker.CHAT_MSG_WHISPER_FILTER)			-- enable received whisper event tracking
	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", Tracker.CHAT_MSG_WHISPER_INFORM_FILTER)	-- enable sent whisper event tracking
	self:RegisterEvent("CHAT_MSG_WHISPER")
	self:RegisterEvent("CHAT_MSG_LOOT")
	HeadCount:GetModule("GUI"):Refresh()
end

-- Processes ending of a raid.
function Tracker:endRaid()
	self.isCurrentRaidActive = false						-- current raid is inactive
	self.isWaitlistAcceptanceEnabled = false				-- waitlist acceptance is disabled
	ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER", Tracker.CHAT_MSG_WHISPER_FILTER)			-- enable received whisper event tracking
	ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER_INFORM", Tracker.CHAT_MSG_WHISPER_INFORM_FILTER)	-- enable sent whisper event tracking
	self:UnregisterEvent("CHAT_MSG_WHISPER")
	self:UnregisterEvent("CHAT_MSG_LOOT")
	self.currentBossEvent:Stop()
	self:CancelAllTimers()
end

-- Determines if there is an active raid.
-- @return boolean Returns true if there is an active raid, and false otherwise.
function Tracker:isRaidActive()
	return self.isCurrentRaidActive and self.raidListWrapper.numberOfRaids > 0
end

function Tracker:endCurrentRaid(activityTime)
	if activityTime and self.isCurrentRaidActive then
		-- end the currently active raid
		if self.raidListWrapper:retrieveMostRecentRaid() then
			self:endRaid()
			-- finalize the current raid
			self.raidListWrapper:retrieveMostRecentRaid():finalize(activityTime)
			return true
		end
	end
	return false
end

function Tracker:removeRaid(id)
	local isRaidRemoved = false
	local isCurrentRaidEnded = false

	local raid = self.raidListWrapper:getRaidById(id)
	if raid then
		-- raid is present and raid
		if not raid.isFinalized and self.isCurrentRaidActive then
			-- raid is not finalized, raid is active
			self:endRaid()
			isCurrentRaidEnded = true
		end
		self.raidListWrapper:removeRaid(id)
		isRaidRemoved = true
	end
	return isRaidRemoved, isCurrentRaidEnded
end
function Tracker:removeAllRaids()
	self:endRaid()
	self.raidListWrapper:removeAll()
end
function Tracker:removePlayer(id, playerName)
	local raid = self.raidListWrapper:getRaidById(id)
	if raid then raid:removePlayer(playerName) end
	self:RAID_ROSTER_UPDATE(true)
end
function Tracker:removeWaitlistPlayer(id, playerName)
	local raid = self.raidListWrapper:getRaidById(id)
	if raid then raid:removeWaitlistPlayer(playerName) end
	self:RAID_ROSTER_UPDATE(true)
end
function Tracker:removeBoss(id, bossName)
	local raid = self.raidListWrapper:getRaidById(id)
	if raid then raid:removeBoss(bossName) end
	self:RAID_ROSTER_UPDATE(true)
end
function Tracker:removeEventAttendee(raidId, bossName, attendeeId)
	local raid = self.raidListWrapper:getRaidById(raidId)
	if raid then raid:removeEventAttendee(bossName, attendeeId)	end
	self:RAID_ROSTER_UPDATE(true)
end

function Tracker:removeLoot(raidId, lootId)
	local raid = self.raidListWrapper:getRaidById(raidId)
	if raid then raid:removeLoot(lootId) end
	self:RAID_ROSTER_UPDATE(true)
end
function Tracker:addRaid(raid)
	local tp = raid:retrieveStartingTime()
	local raidStartTime = tp.epoc
	self.raidListWrapper:addRaid(raidStartTime, raid)
	self:saveRaidListWrapper()
	HeadCount:GetModule("GUI"):Refresh()
end
function Tracker:retrieveCurrentRaid(activityTime)
	local currentRaid = nil
	local isNewRaid = false
	if self.isCurrentRaidActive then
		-- a raid currently exists, process existing raid
		currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
		isNewRaid = false
	else
		-- Check for a recover
		self:recover()
		if not self.isCurrentRaidActive then
			-- a raid does not already exist, create a new raid, then process it
			local timePair = HeadCount:NewInstance("TimePair")
			timePair.start = activityTime
			local args = {
				["playerList"] = {
					["raidgroup"] = { },
					["nogroup"] = { },
				},
				["bossList"] = { },
				["lootList"] = { },
				["timeList"] = {
					[1] = timePair
				},
				["difficulty"] = 1,
				["raidTime"] = 0,
				["lastActivityTime"] = activityTime,
				["zone"] = UNKNOWN,
				["numberOfPlayers"] = 0,
				["numberOfBosses"] = 0,
				["isFinalized"] = false,
				["lastBoss"] = HeadCount.DEFAULT_LOOT_SOURCE,
			}
			currentRaid = HeadCount:NewInstance("Raid",args)
			self:addRaid(currentRaid)
			isNewRaid = true
			HeadCount:LogInformation(string.format(L["info.newraid"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(currentRaid.lastActivityTime)))
			-- Set proper events and flags for new raid
			self:beginRaid()
		else
			currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
		end
	end
	return currentRaid, isNewRaid
end

function Tracker:processZoneChange()
	local zone = GetRealZoneText()
	-- update the status
	self:processStatus(zone)
	if (self.isEnabled) then
		-- check if the current zone is a valid raid instance zone
		self:processAutomaticGroupSelection(zone)	-- process automatic group selection if applicable
		self:processRaidZone(zone)					-- process the zone
		self:processDifficulty()					-- process the instance difficulty
	end
	return zone
end

function Tracker:processStatus(zone)
	if not zone then return end
	local inInstance, instanceType = IsInInstance()
	if not HeadCount.db.profile.isBattlegroundTrackingEnabled and HeadCount:isBattlegroundZone(zone) then
		-- battleground tracking is disabled and current zone is a battleground
		if self.isEnabled then
			HeadCount:LogDebug(string.format(L["debug.status.battleground.enter"], HeadCount.TITLE, HeadCount.VERSION, zone))
			self.isEnabled = false
		end
	elseif not self.isEnabled then
		if instanceType == "raid" then
			self.isEnabled = true
		end
		if HeadCount:isBattlegroundZone(zone)then
			HeadCount:LogDebug(string.format(L["debug.status.battleground.leave"], HeadCount.TITLE, HeadCount.VERSION))
			self.isEnabled = false
		end
	end
end


function Tracker:processAutomaticGroupSelection(zone)
	local isAutomaticGroupSelectionEnabled = HeadCount.db.profile.raid.groupsetup.isAutomaticGroupSelectionEnabled
	local isPresentInInstance, instanceType = IsInInstance()
	if isPresentInInstance and instanceType == "raid" and self.isCurrentRaidActive and isAutomaticGroupSelectionEnabled then
		local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
		if type == "raid" then
			local numberOfPlayers = maxPlayers
			local numberOfRaidGroups = numberOfPlayers / HeadCount.NUMBER_OF_PARTY_PLAYERS
			for i = 1, HeadCount.NUMBER_OF_PARTY_GROUPS do
				-- 10 / 5 = 2,  20 / 5 = 4,  25 / 5 = 5, 40 / 5 = 8
				local partyGroupString = string.format("%d", i)
				if (i <= numberOfRaidGroups) then
					HeadCount.db.profile.raidListGroups[partyGroupString] = true
					HeadCount.db.profile.waitListGroups[partyGroupString] = false
				else
					HeadCount.db.profile.raidListGroups[partyGroupString] = false
					HeadCount.db.profile.waitListGroups[partyGroupString] = true
				end
			end
		end
	end
end

-- Processes the current raid zone.
-- If the current raid has no raid zone assigned to it, set the raid zone name.
-- @param zone The zone name.
function Tracker:processRaidZone(zone)
	local isPresentInInstance, instanceType = IsInInstance()
	if isPresentInInstance and instanceType == "raid" and self.isCurrentRaidActive then
		-- player is in a raid instance and a tracked raid is active
		local currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
		local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
		if not currentRaid.zone or currentRaid.zone == UNKNOWN  then
			-- current raid zone is not set
			local localizedZoneName = name
			currentRaid.zone = localizedZoneName
			currentRaid.difficulty = HeadCount:determineDifficulty()
			self:saveRaidListWrapper()	-- save the raid list
		end
	end
end

-- Processes the difficulty
function Tracker:processDifficulty()
	local isPresentInInstance, instanceType = IsInInstance()
	if isPresentInInstance and instanceType == "raid" and self.isCurrentRaidActive then
		-- player is in a raid instance and a tracked raid is active
		local currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
		if not currentRaid.difficulty then
			local difficulty = HeadCount:determineDifficulty()
			local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
			if difficulty then
				HeadCount:LogDebug(string.format(L["debug.raid.update.difficulty"], HeadCount.TITLE, HeadCount.VERSION, difficultyName))
				currentRaid.difficulty = difficulty
				self:saveRaidListWrapper()	-- save the raid list
			end
		end
	end
end


function Tracker:processLootUpdate(message)
	if self.isCurrentRaidActive then
		-- current raid is active
		local playerName, item, quantity = self:processLootPattern(message)
		if playerName then
			-- message matched proper loot pattern
			local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
			local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)
			local lootSource = self:processLootSource()
			-- Process the loot
			self:processLoot(playerName, item, quantity, activityTime, lootSource)
		end
	end
end

-- Process the loot source
-- @return isBoss Returns if the loot source is a boss
-- @return lootSource Returns the loot source (mob name)
function Tracker:processLootSource()
	-- if we are recovering, we need to figure out last boss killed
	return self.currentBossEvent.lastBoss
end

-- Processing a loot
-- @param playerName The player name.
-- @param item The item string
-- @param quantity The loot quantity
-- @param activityTime The activity time.
-- @param source The loot source
-- @return boolean Returns true if loot was processed and false otherwise
function Tracker:processLoot(playerName, item, quantity, activityTime, source)
	local isProcessed = false
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item)

	local minimumLootQuality = HeadCount.db.profile.minimumLootQuality
	local id = string.match(itemLink, "item:(%d+):")
	local numberId = tonumber(id)
	local exclusionList = HeadCount:GetExclusionList()
	if itemRarity >= minimumLootQuality and not exclusionList[numberId] then
		-- item dropped is greater than or equal to minimum loot quality and the item is not excluded, track it!
		local realZoneText = GetRealZoneText()
		local args = {
			["player"] = playerName,
			["cost"] = 0,
			["looted"] = activityTime,
			["quantity"] = quantity,
			["zone"] = realZoneText,
			["source"] = source,
			["link"] = itemLink,
			["note"] = nil
		}
		--local itemId = AceLibrary("HeadCountLoot-1.0"):retrieveItemId(itemLink)	-- get the item id
		local currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
		local loot = HeadCount:NewInstance("Loot", args)

		-- Add this loot to the raid
		currentRaid:addLoot(loot)

		-- Should we broadcast this loot to guild chat?
		if HeadCount.db.profile.isLootBroadcastEnabled then
			-- player is in a guild and loot broadcasting is enabled, broadcast to the guild!
			local channel = HeadCount.db.profile.broadcastChannel
			HeadCount:SendMessage(string.format(L["guild.broadcast.loot"], HeadCount.TITLE, HeadCount.VERSION, playerName, itemLink, source), channel)
		end

		-- Should the loot management popup window show?
		if HeadCount.db.profile.isLootPopupEnabled then
			-- loot popup window should show
			local raidId = currentRaid:retrieveStartingTime().epoc
			local lootId = #currentRaid.lootList
			HeadCount:GetModule("GUI"):ShowLootPopup(raidId, lootId)
		end
		isProcessed = true
		self:saveRaidListWrapper()
		HeadCount:GetModule("GUI"):RefreshActivePanel()
	end
	return isProcessed
end

-- Processes a loot message pattern
-- @param message The loot message.
-- @return string, string, number  Returns the player name, item name, and quantity or nil if the message does not match any pattern
function Tracker:processLootPattern(message)
	local playerName = nil
	local item = nil
	local quantity = nil

	local multipleLootSelfRegex = HeadCount:getMULTIPLE_LOOT_SELF_REGEX()
	local lootSelfRegex = HeadCount:getLOOT_SELF_REGEX()
	local multipleLootRegex = HeadCount:getMULTIPLE_LOOT_REGEX()
	local lootRegex = HeadCount:getLOOT_REGEX()

	if (message:match(multipleLootSelfRegex)) then
		item, quantity = message:match(multipleLootSelfRegex)
		playerName = UnitName("player")
	elseif (message:match(lootSelfRegex)) then
		item = message:match(lootSelfRegex)
		playerName = UnitName("player")
		quantity = 1
	elseif (message:match(multipleLootRegex)) then
		playerName, item, quantity = message:match(multipleLootRegex)
	elseif (message:match(lootRegex)) then
		playerName, item = message:match(lootRegex)
		quantity = 1
	end
	return playerName, item, quantity
end

-- Add a boss kill.
-- @param encounterName The boss encounter name.
-- @param zone The zone name.
-- @param activityTime The boss kill time.
-- @return boolean Returns true if a boss kill was processed and false otherwise
function Tracker:addBossKill(encounterName, zone, difficulty, activityTime)
	local isProcessed = false
	local currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
	if not currentRaid:retrieveBoss(encounterName) then
		HeadCount:LogDebug(string.format(L["info.boss.kill.complete"], HeadCount.TITLE, HeadCount.VERSION, encounterName))

		self:processAttendance(activityTime)	-- update attendance

		local raidMemberSort = HeadCount:getRAID_MEMBER_SORT()
		local playerList = currentRaid:retrieveOrderedPlayerList("Name",
			HeadCount.db.profile.boss["groupsetup"]["raidlistgroup"], HeadCount.db.profile.boss["groupsetup"]["waitlistgroup"],	HeadCount.db.profile.boss["groupsetup"]["nolistgroup"],
			HeadCount.db.profile.boss["groupsetup"]["waitlist"], true, true)
		local args = {
			["name"] = encounterName,
			["zone"] = zone,
			["difficulty"] = difficulty,
			["runtime"] = activityTime,
			["players"] = playerList
		}
		local boss = HeadCount:NewInstance("Boss",args)
		currentRaid:addBoss(boss)
		-- Should we broadcast this to guild chat?
		if HeadCount.db.profile.isBossBroadcastEnabled then
			-- boss kill broadcasting is enabled
			local channel =  HeadCount.db.profile.broadcastChannel
			HeadCount:SendMessage(string.format(L["guild.broadcast.bosskill"], HeadCount.TITLE, HeadCount.VERSION, encounterName), channel)
		end
		self:saveRaidListWrapper()
		HeadCount:GetModule("GUI"):RefreshActivePanel()
	end
end

-------------------------------
-- ATTENDANCE TRACKING
-------------------------------
-- Process an attendance update
-- @param activityTime The current activity time.
function Tracker:processAttendance(activityTime)
	assert(activityTime, "Unable to update attendance because the activity time is nil.")

	-- Get the current zone
	local zone = GetRealZoneText()
	-- update the status
	self:processStatus(zone)
	local numberOfRaidMembers = GetNumRaidMembers()
	if self.isEnabled then
		if numberOfRaidMembers > 0 then
			-- player is in a raid
			local currentRaid, isNewRaid = self:retrieveCurrentRaid(activityTime)	-- retrieve the current raid or create a new raid
			if isNewRaid then
				-- new raid Hide anything we need to hide
			end
			if not currentRaid then
				HeadCount:LogDebug("Critical Failure, Unableto fetch raid record")
				return
			end
			if not currentRaid.zone or currentRaid.zone == UNKNONW then
				-- process automatic group selection if enabled and only if no zone is already defined for this raid
				self:processAutomaticGroupSelection(zone)
			end
			self:processRaidZone(zone)			-- Assign zone if current raid does not have a zone assigned
			self:processDifficulty()			-- Assign difficulty is current difficulty is not yet assigned

			-- Update each member in the raid group

			if numberOfRaidMembers > 0 then
				-- active in a raid

				local trackedPlayerList = { }
				for i = 1,numberOfRaidMembers do
					local name, rank, subgroup, level, className, fileName, zone, isOnline, isDead, role, isML = GetRaidRosterInfo(i)
					subgroup = tostring(subgroup)
					if currentRaid:isPlayerPresent(name) then
						-- player is already present in the player list (existing player)
						local player = currentRaid:retrievePlayer(name)
						local timeDifference = HeadCount:computeTimeDifference(activityTime, player.lastActivityTime)
						local moveResult

						if isOnline then
							-- CURRENT: Player is online
							local isPresentInRaidList = HeadCount.db.profile.raidListGroups[subgroup]
							local isPresentInWaitList = HeadCount.db.profile.waitListGroups[subgroup]

							if isPresentInRaidList then
								-- CURRENT: Player is in the raid list.
								moveResult = player:moveToRaidList(activityTime)
								if moveResult then
									currentRaid:moveToRaidGroup(player)
									HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.raidlist"], HeadCount.TITLE, HeadCount.VERSION, name))
								end
							elseif isPresentInWaitList then
								-- CURRENT: Player is in the wait list.
								moveResult = player:moveToWaitList(activityTime)
								if moveResult then
									currentRaid:moveToRaidGroup(player)
									HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.waitlist"], HeadCount.TITLE, HeadCount.VERSION, name))
								end
							else
								moveResult = player:moveToNoList(activityTime, true)
								if moveResult then
									currentRaid:moveToNoGroup(player)
									HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.nolist"], HeadCount.TITLE, HeadCount.VERSION, name))
								end
							end
						else
							-- CURRENT: Player is offline
							moveResult = player:moveToNoList(activityTime, false)
							if moveResult then
								currentRaid:moveToNoGroup(player)
								HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.offline"], HeadCount.TITLE, HeadCount.VERSION, name))
							end
						end

						player:updateValues(i, className, fileName)	-- update player values if needed
					else
						-- CREATE NEW PLAYER
						-- player is NOT present in the player list (new player)
						local _, race = UnitRace("raid" .. i)
						local sex = UnitSex("raid" .. i)
						local guild = GetGuildInfo("raid" .. i)

						if 0 == level then
							level = UnitLevel("raid" .. i)
						end

						if name then
							-- protect against phantom players
							local player = self:createPlayer(name, race, guild, sex, level, className, fileName, subgroup, isOnline, activityTime)
							currentRaid:addPlayer(player)	-- add the player to the raid player list
						end
					end	-- end raid member is existing present member or new member

					-- track the player
					if name then
						trackedPlayerList[name] = name
					end
				end	-- end loop through all raid members

				-- TODO: Remove/cleanup members that are no longer in the raid.
				currentRaid:processMissingPlayers(numberOfRaidMembers, trackedPlayerList, activityTime)

				-- Update the last activity time for the raid after member processing is fully complete
				currentRaid.lastActivityTime = activityTime
				HeadCount:LogDebug(string.format(L["debug.raid.update.continue"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
				HeadCount:LogDebug(string.format(L["debug.raid.update.numberOfRaidMembers"], HeadCount.TITLE, HeadCount.VERSION, currentRaid:GetNumberOfPlayers(), HeadCount:getDateTimeAsString(activityTime)))
			else
				-- no raid active, cleanup any existing raids
				if self.isCurrentRaidActive then
					-- current raid is active, end it
					HeadCount:LogDebug(string.format(L["debug.raid.update.activitytime"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
					self:endCurrentRaid(activityTime)
					HeadCount:LogDebug(string.format(L["debug.raid.update.end"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
				end
			end	-- end determine number of raid members
		else
			-- cleanup any existing raids
			if self.isCurrentRaidActive then
				-- current raid is active, end it
				HeadCount:LogDebug(string.format(L["debug.raid.update.activitytime"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
				self:endCurrentRaid(activityTime)
				HeadCount:LogDebug(string.format(L["debug.raid.update.end"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
			end
		end -- end determine if mod owner is in a raid or not
		--self:SetRaidListWrapper(self.raidListWrapper)	-- save the raid list
	end
end

-- Manage loot updates
-- @param raidId The raid id.
-- @param lootId The loot id.
-- @param playerName The name of the player who looted the item
-- @paam source The loot source
-- @param lootCost The loot cost.
-- @param lootNote The loot note.
function Tracker:lootManagementUpdate(raidId, lootId, playerName, source, lootCost, lootNote)
	assert(raidId, "Unable to process loot management update because the raidId is not found.")
	assert(lootId, "Unable to process loot management update because the lootId is not found.")
	assert(playerName, "Unable to process loot management update because the player name is nil.")
	assert(source, "Unable to process loot management update because the loot source is nil.")
	assert(type(lootCost) == "number", "Unable to process loot management update because the loot cost is not a number.")
	assert(lootNote, "Unable to process loot management update because the loot note is nil.")

	local raid = self.raidListWrapper:getRaidById(raidId)
	if raid then
		local loot = raid:retrieveLoot(lootId)
		if loot then
			loot.player = playerName
			loot.source = source
			loot.cost = lootCost
			loot.note = lootNote
			--self:SetRaidListWrapper(self.raidListWrapper)	-- save the raid list
		else
			error("Unable to process loot management update because the loot could not be found: " .. lootId)
		end
	else
		error("Unable to process loot management update because the raid could not be found: " .. raidId)
	end
end

-- Creates a player.
-- @param name The name.
-- @param race The race.
-- @param guild The guild name.
-- @param sex The sex.
-- @param level The player level.
-- @param className The localized class name.
-- @param fileName The standard class name.
-- @param subgroup The subgroup.
-- @param isOnline The online status.
-- @param activityTime The activity time.
-- @return object Returns a player or nil if none can be created.
function Tracker:createPlayer(name, race, guild, sex, level, className, fileName, subgroup, isOnline, activityTime)
	local player = nil

	local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
	local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)

	local timePair = HeadCount:NewInstance("TimePair", { start = activityTime})

	local args = {
		["name"] = name,
		["className"] = className,
		["fileName"] = fileName,
		["race"] = race,
		["guild"] = guild,
		["sex"] = sex,
		["level"] = level,
		["isPresentInRaidList"] = false,
		["isPresentInWaitList"] = false,
		["isOnline"] = true,
		["isWaitlisted"] = false,
		["waitlistActivityTime"] = nil,
		["waitlistNote"] = nil,
		["raidListTime"] = 0,
		["waitListTime"] = 0,
		["offlineTime"] = 0,
		["timeList"] = { },
		["lastActivityTime"] = activityTime,
		["isFinalized"] = false
	}

	if isOnline then
		-- CURRENT: Player is online
		if HeadCount.db.profile.raidListGroups[subgroup] then
			-- CURRENT: Player is in the raid list.
			args["isPresentInRaidList"] = true
			args["isPresentInWaitList"] = false
			args["isOnline"] = true

			timePair.note = L["Raid list"]
			table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

			player = HeadCount:NewInstance("Player",args)
			HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.raidlist"], HeadCount.TITLE, HeadCount.VERSION, name))
		elseif HeadCount.db.profile.waitListGroups[subgroup] then
			-- CURRENT: Player is in the wait list.
			args["isPresentInRaidList"] = false
			args["isPresentInWaitList"] = true
			args["isWaitlisted"] = true
			args["waitlistActivityTime"] = activityTime
			args["isOnline"] = true

			timePair.note = L["Wait list"]
			table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

			player = HeadCount:NewInstance("Player",args)
			HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.waitlist"], HeadCount.TITLE, HeadCount.VERSION, name))
		else
			-- CURRENT: Player is in no list.
			args["isPresentInRaidList"] = false
			args["isPresentInWaitList"] = false
			args["isOnline"] = true

			timePair.note = L["No list"]
			table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

			player = HeadCount:NewInstance("Player",args)
			HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.nolist"], HeadCount.TITLE, HeadCount.VERSION, name))
		end
	else
		-- CURRENT: Player is offline
		args["isPresentInRaidList"] = false
		args["isPresentInWaitList"] = false
		args["isOnline"] = false

		timePair.note = L["Offline"]
		table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

		player = HeadCount:NewInstance("Player",args)
		HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.offline"], HeadCount.TITLE, HeadCount.VERSION, name))
	end
	HeadCount:GetModule("GUI"):Refresh()
	return player
end

----------------------------
-- BOSS KILL TRACKING
----------------------------

-- Prunes old raids.
function Tracker:pruneRaids()
	local isPruningEnabled = HeadCount.db.profile.raid["pruning"]["isPruningEnabled"]
	local pruningTimeInWeeks = HeadCount.db.profile.raid["pruning"]["pruningTime"]

	if isPruningEnabled then
		HeadCount:LogDebug(string.format(L["debug.raid.prune.enabled"], HeadCount.TITLE, HeadCount.VERSION, pruningTimeInWeeks))
		self.raidListWrapper:pruneRaids(pruningTimeInWeeks)	-- prune old raids
	else
		HeadCount:LogDebug(string.format(L["debug.raid.prune.disabled"], HeadCount.TITLE, HeadCount.VERSION))
	end
end

-- Processes an incoming message.
-- @param message The message.
-- @param author The author.
-- @return boolean Returns true if an update is required as a result of the message.
-- @return boolean Returns true if the message is a supported message.
-- @return table Returns the response message list.
function Tracker:processIncomingMessage(message, author)
	local isUpdateRequired = false
	local isSupportedMessage = false
	local messageList = nil
	if self.isCurrentRaidActive then
		-- Only process incoming messages during an active raid
		local raid = self:retrieveCurrentRaid()
		isUpdateRequired, isSupportedMessage, messageList = Tracker.messageHandler:processIncomingMessage(raid, self.isWaitlistAcceptanceEnabled, message, author)
	end
	return isUpdateRequired, isSupportedMessage, messageList
end

-- Determines if an incoming message is supported.
-- @param message The message.
-- @return boolean Returns true if the message is a supported message.
function Tracker:IsSupportedIncomingMessage(message)
	assert(message, "Unable to determine if incoming message is supported because the message is nil.")
	local isSupportedMessage = false
	if self.isCurrentRaidActive then
		isSupportedMessage = Tracker.messageHandler:IsSupportedIncomingMessage(message)
	end
	return isSupportedMessage
end

-- Determines if an outgoing message is supported.
-- @param message The message.
-- @return boolean Returns true if the message is a supported message.
function Tracker:IsSupportedOutgoingMessage(message)
	assert(message, "Unable to determine if the outgoing message is supported because the message is nil.")
	local isSupportedMessage = false
	if self.isCurrentRaidActive then
		isSupportedMessage = Tracker.messageHandler:IsSupportedOutgoingMessage(message)
	end
	return isSupportedMessage
end

function Tracker:processListWaitlist(channel)
	if self.isCurrentRaidActive then
		local raid = self:retrieveCurrentRaid()
		Tracker.messageHandler:processListWaitlistMessage(raid, channel)
	else
		HeadCount:LogWarning(string.format(L["warning.waitlist.list.currentraid"], HeadCount.TITLE, HeadCount.VERSION))
	end
end

local whisperData = { }
local evtTimer = nil
--[[
	Event Handlers
--]]
function Tracker:CHAT_MSG_WHISPER_FILTER(chatFrame,event,...)
	local message = ...
	if Tracker.messageHandler:IsSupportedIncomingMessage(message) then
		-- waitlist message and author was not self
		if not HeadCount.db.profile.raid["waitlist"]["incoming"] then
			-- waitlist incoming message display is disabled
			return true, ...
		end
	end
end
function Tracker:CHAT_MSG_WHISPER_INFORM_FILTER(chatFrame,event,...)
	local message = ...
	if Tracker.messageHandler:IsSupportedOutgoingMessage(message) and not HeadCount.db.profile.raid["waitlist"]["incoming"] then
		-- outgoing message is a waitlist message and outgoing waitlist message display is disabled
		return true, ...
	end
end
function Tracker:OUTGOING_WHISPER_UPDATE()
	local numberOfMessages = # whisperData
	if (numberOfMessages > 0) then
		local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
		local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)

		-- message is ready to be sent
		local messages = whisperData[1]["messages"]
		local recipient = whisperData[1]["recipient"]

		if ((messages) and (recipient)) then
			for k,v in ipairs(messages) do
				HeadCount:SendWhisper(v, recipient)
			end

			table.remove(whisperData, 1)
		else
			table.remove(whisperData, 1)
		end

		local updatedNumberOfMessages = # whisperData
		if (updatedNumberOfMessages > 0) then
			-- more messages to send
			self:CancelAllTimers()
			evtTimer = self:ScheduleTimer(Tracker.OUTGOING_WHISPER_UPDATE,HeadCount.EVENT_OUTGOING_WHISPER_UPDATE_DELAY,self)
		else
			HeadCount:LogDebug(string.format(L["debug.whisper.nomessages"], HeadCount.TITLE, HeadCount.VERSION))
		end
	end
end

function Tracker:CHAT_MSG_WHISPER(event, message,author, ...)
	local isUpdateRequired, isSupportedMessage, messageList = self:processIncomingMessage(message, author)
	if isUpdateRequired then
		Tracker:RAID_ROSTER_UPDATE(false)	-- update attendance and the UI
	end

	if isSupportedMessage then
		-- waitlist message and author was not self
		-- send back response
		HeadCount:LogDebug(string.format(L["debug.message"], HeadCount.TITLE, HeadCount.VERSION, message, author))
		table.insert(whisperData, { ["messages"] = messageList, ["recipient"] = author, })
		evtTimer = self:ScheduleTimer(Tracker.OUTGOING_WHISPER_UPDATE,HeadCount.EVENT_OUTGOING_WHISPER_UPDATE_DELAY,self)
		if not HeadCount.db.profile.raid.waitlist.incoming then
			-- waitlist incoming message display is disabled
			return true, ...
		end
	end
end

local wlTimer = nil
-- Announce the wait list
-- @param channel The chat channel.
function Tracker:announceWaitlist(channel)
	if self.isCurrentRaidActive then
		self.isWaitlistAcceptanceEnabled = true
        -- Always cancel the schedule event if one exists
        if self:TimeLeft(wlTimer) then
			self:CanelceTimer(wlTimer,true)
		end

        local waitlistDuration = HeadCount.db.profile.raid["waitlist"]["duration"]
        if waitlistDuration > 0 then
			local waitlistDurationSeconds = waitlistDuration * 60		-- convert to seconds
			wlTimer = self:ScheduleTimer(Tracker.WAITLIST_ACCEPTANCE_UPDATE,waitlistDurationSeconds,self.channel)
		end
		HeadCount:SendMessage(HeadCount.MESSAGE_PREFIX .. " " .. L["info.announce.waitlist.open"], channel)
	else
		HeadCount:LogWarning(string.format(L["warning.waitlist.announce.currentraid"], HeadCount.TITLE, HeadCount.VERSION))
	end
end

-- Automatically called when the wait list acceptance timer expires
function Tracker:WAITLIST_ACCEPTANCE_UPDATE(channel)
	self:CancelTimer(wlTimer,true)
    local waitlistDuration = HeadCount.db.profile.raid["waitlist"]["duration"]
    self:manageWaitlistChange(duration)
    if (waitlistDuration > 0) then
		-- Display closed message if duration is non-zero
		HeadCount:SendMessage(HeadCount.MESSAGE_PREFIX .. " " .. L["info.announce.waitlist.close"], channel)
	end
end

-- Manages a wait list status change
function Tracker:manageWaitlistChange(duration)
	if self.isCurrentRaidActive then
		-- raid is active
		if self:TimeLeft(wlTimer) then
			if (0 == duration) then
				-- wait list status is enabled since duration is 0
				self.isWaitlistAcceptanceEnabled = true
			end
		else
			-- no wait list event is currently scheduled so it is ok to change the current waitlist status
			if (0 == duration) then
				-- wait list status is enabled since duration is 0
				self.isWaitlistAcceptanceEnabled = true
			else
				-- wait list status is enabled
				self.isWaitlistAcceptanceEnabled = false
			end
		end
	else
		-- raid is not active
		-- disable the wait list
		self.isWaitlistAcceptanceEnabled = false
	end
end


function Tracker:RAID_ROSTER_UPDATE(forced)
	local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
	local activityTime = HeadCount:NewInstance("Time", utcDateTimeInSeconds)
	--HeadCount:LogDebug("Forced update?: " .. HeadCount:convertBooleanToString(booleanValue))
	local numberOfRaidMembers = GetNumRaidMembers()

	if numberOfRaidMembers > 0 and self.isCurrentRaidActive and not forced then
		HeadCount:LogInformation("Raid is currently active.")
		local raid = Tracker.raidListWrapper:retrieveMostRecentRaid()
		local lastActivityTime = raid.lastActivityTime
		local timeDifference = HeadCount:computeTimeDifference(activityTime, lastActivityTime)
		local delay = HeadCount.EVENT_OUTGOING_WHISPER_UPDATE_DELAY

		if (timeDifference >= delay) then
			self:processAttendance(activityTime) -- update the attendance
			--update the ui
		else
			self:ScheduleTimer(Tracker.RAID_ROSTER_UPDATE,(delay - timeDifference),self)
		end

	else
		--HeadCount:LogInformation("Raid is currently inactive or a forced raid update is occurring.  Processing attendance.")
		self:processAttendance(activityTime)	-- update the raid attendance
		-- update the gui
	end
end

function Tracker:CHAT_MSG_LOOT(event, message)
	local isProcessed = self:processLootUpdate(message)
	if isProcessed then
		self:RAID_ROSTER_UPDATE(true)	-- update attendance and the UI immediately
	end
end

function Tracker:BossKilled(event,boss)
	local guid = nil
	if not boss then return end
	HeadCount:LogDebug("Boss Killed "..boss)

	if self.isCurrentRaidActive then
		if self.currentBossEvent.started then
			guid = self.currentBossEvent:GetGUID(boss)
			if not guid then
				-- we are told a boss died but we didnt see the engage message
				-- let fake it up
				local zone = GetRealZoneText()
				local difficulty = HeadCount:determineDifficulty()
				self.currentBossEvent:AddBoss(guid,boss,zone,difficulty)
			end
			local isPresent = self.currentBossEvent:IsPresent(guid)
			local isAlive = self.currentBossEvent:IsAlive(guid)
			if isPresent and isAlive then
				-- boss kill recorded!
				HeadCount:LogInformation(string.format(L["info.boss.kill"], HeadCount.TITLE, HeadCount.VERSION, self.currentBossEvent:GetName(guid)))
				self.currentBossEvent:AddKill(guid)
				if (self.currentBossEvent:Complete()) then
					-- boss event is complete!
					local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
					local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)
					local zone = GetRealZoneText()
					local difficulty = HeadCount:determineDifficulty()
					self:addBossKill(boss, zone, difficulty, activityTime)	-- add the boss kill
					self.currentBossEvent:Stop()
				end
			end
		else
			-- We werent in an active boss setup means we didnt see the engage
			-- let fake it up
			local guid = HeadCount:GenerateID(boss)
			local zone = GetRealZoneText()
			local difficulty = HeadCount:determineDifficulty()
			self.currentBossEvent:AddBoss(guid,boss,zone,difficulty)
			self.currentBossEvent:AddKill(guid)
			HeadCount:LogInformation(string.format(L["info.boss.kill"], HeadCount.TITLE, HeadCount.VERSION, self.currentBossEvent:GetName(guid)))
			local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
			local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)
			local zone = GetRealZoneText()
			local difficulty = HeadCount:determineDifficulty()
			self:addBossKill(boss, zone, difficulty, activityTime)	-- add the boss kill
			self.currentBossEvent:Stop()
		end
	else
		self:endRaid()
	end
end

function Tracker:BossWipe(event,boss)
	HeadCount:LogDebug("Boss Wiped "..boss)

	if self.isCurrentRaidActive then
		if self.currentBossEvent.started then
			HeadCount:LogInformation(string.format(L["info.boss.wipe"], HeadCount.TITLE, HeadCount.VERSION,boss))
			self.currentBossEvent:Stop(true)
		end
	else
		self:endRaid()
	end
end

function Tracker:BossEngaged(event,boss)
	HeadCount:LogDebug("Boss Engaged "..boss)
	if self.isCurrentRaidActive then
		local guid = HeadCount:GenerateID(boss)
		if not self.currentBossEvent:IsPresent(guid) then
			local zone = GetRealZoneText()
			local difficulty = HeadCount:determineDifficulty()
			HeadCount:LogInformation(string.format(L["info.boss.engage"], HeadCount.TITLE, HeadCount.VERSION,boss))
			self.currentBossEvent:AddBoss(guid,boss,zone,difficulty)
		else
			self.currentBossEvent.started = true
			HeadCount:LogInformation(string.format(L["info.boss.engage"], HeadCount.TITLE, HeadCount.VERSION,boss))
		end
	else
		self:endRaid()
	end
end


function Tracker:retrieveAttendanceTotals(sortColumn, descending)

	local allRaids = self.raidListWrapper:retrieveOrderedRaidList(false)

	local totalRaids = #allRaids
	local thirtyCount = 0
	local sixtyCount = 0
	local ninetyCount = 0

	local dataSet = {}
	local now = HeadCount:getUTCDateTimeInSeconds()
	-- How many raid do we have?
	for k,v in pairs(allRaids) do
		-- Get the raids Start Time
		local raidStart = v:retrieveStartingTime().epoc
		local diff = difftime(now, raidStart)
		-- Figure out totals for each subset
		if diff <= HeadCount.THIRTY_DAYS then
			thirtyCount = thirtyCount + 1
		elseif diff <= HeadCount.SIXTY_DAYS then
			sixtyCount = sixyCount + 1
		else
			ninetyCount = ninetyCount + 1
		end

		local players = v:retrieveOrderedPlayerList("Name",true,true,true,true,true,true)
		-- scan all players found
		for pI, playerName in pairs(players) do
			if not dataSet[playerName] then
				dataSet[playerName] = {
					["thirty"] = 0,
					["sixty"] = 0,
					["ninety"] = 0,
					["total"] = 0,
					["name"] = playerName
				}
			end
			if diff <= HeadCount.THIRTY_DAYS then
				dataSet[playerName].thirty = dataSet[playerName].total + 1
			elseif diff <= HeadCount.SIXTY_DAYS then
				dataSet[playerName].sixty = dataSet[playerName].total + 1
			else
				dataSet[playerName].ninety = dataSet[playerName].total + 1
			end
			dataSet[playerName].total = dataSet[playerName].total + 1
		end
	end
	-- How many raids was a given player present in those raids
	local sortedList = {}
	for k,v in pairs(dataSet) do
		v.thirty = v.thirty/thirtyCount * 100
		if thirtyCount == 0 then
			v.thirty = -1
		end
		v.sixty = v.sixty/sixtyCount * 100
		if sixtyCount == 0 then
			v.sixty = -1
		end
		v.ninety = v.ninety/ninetyCount * 100
		if ninetyCount == 0 then
			v.ninety = -1
		end
		v.total = v.total/totalRaids * 100
		if totalRaids == 0 then
			v.total = -1
		end
		table.insert(sortedList,v)
	end
	-- Ok build list and sort
	if sortColumn == "Name" then
		table.sort(sortedList, function(a,b)
			if descending then
				return a.name < b.name
			else
				return a.name > b.name
			end
		end)
	end
	if sortColumn == "Thirty" then
		table.sort(sortedList, function(a,b)
			if descending then
				return a.thirty < b.thirty
			else
				return a.thirty > b.thirty
			end
		end)
	end
	if sortColumn == "Sixety" then
		table.sort(sortedList, function(a,b)
			if descending then
				return a.sixty < b.sixty
			else
				return a.sixty > b.sixty
			end
		end)
	end
	if sortColumn == "Ninety" then
		table.sort(sortedList, function(a,b)
			if descending then
				return a.ninety < b.ninety
			else
				return a.ninety > b.ninety
			end
		end)
	end
	if sortColumn == "Total" then
		table.sort(sortedList, function(a,b)
			if descending then
				return a.total < b.total
			else
				return a.total > b.total
			end
		end)
	end

	return sortedList
end